REBOL [
	Title: "REBOL/Services - Async HTTP Client Handler"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005, All rights reserved."
]

;-- Asynchronous HTTP Client Handler -----------------------------------------
;
; NOTE:
;   This code uses HTTP/1.0. It does not really need the chunked
;   transfer mode of HTTP/1.1. We can do this because the length of
;   a RIS packet is encoded in its header near the beginning of the
;   transfer. The LOAD-PACKET function scans for the RIS header get
;   the length, so no HTTP content length info is needed.

add-client-handler [

	scheme: 'http

	opener: func [
		spec [port!]
		config
		/local session
	][
		debug connect ['opening-http-port spec/host spec/port-id]
		; Build new TCP port based on HTTP port.
		; Let the original HTTP port get recycled.
		if config [
			spec/handler: spec/proxy: none ; help out GC
			spec: make port! [
				scheme: 'tcp
				host: spec/host
				port-id: spec/port-id
				path: spec/path
				target: spec/target
			]
			; Create a session and open the port:
			session: make-session spec config
		]
		open/direct/binary/async spec :handler ; also returns port
	]

	sender: func [
		port [port!]
		request
	][
		request: mold-packet port/user-data request
		make-http-request port request
		insert port request
	]

	set-wakeup: func [
		port [port!]
		flag [logic!]
	] [
		port/wakeup?: any [port/wakeup? flag]
	]

	clear-wakeup: func [
		port [port!]
	] [
		port/wakeup?: none
	]

	handler: func [
		port action arg
		/local session buf
	][
		debug tcp ['handle-http action either object? arg ['object][arg]]
		session: port/user-data
		buf: session/recv-buf
		switch action [
			write-done [
				; Top item in queue is always the one just sent
				if arg: session/send-queue/1 [append session/sent-queue arg]
				remove session/send-queue
				send-request port
			]
			read [
				append buf copy/part port arg
				debug tcp-read [arg 'bytes length? buf]
				while [ ; multiple requests may be concatenated in the buffer
					all [
						not empty? buf
						arg: load-packet session
					]
				][
					wake-request port arg ; arg is block (ok) or word (failure)
				] 
				; Deal with possible HTTP server errors...
				if not arg [ ; No packet, what's up?
					buf: copy-line as-string buf ; server reply
					if not find buf "200" [ ; if not 200, then problem
						debug error ['bad-cgi buf] ; server reply
						;print as-string session/recv-buf ; !!!
						clear session/recv-buf
						wake-request port 'http-server-error
					]
				]
			]
			init [
				port/status: none
				set-modes port [timeout: session/port-timeout]
			]
			open [
				debug connect ['open port/remote-ip port/remote-port port/local-port]
				port/status: 'open
				send-request port
			]
			close [
				debug connect ['close session/id port/remote-ip port/remote-port port/local-port]
				attempt [close port]
				port/status: 'close
				; Check to see if there are any other pending requests:
				if not empty? port/user-data/send-queue [
					reset-async-port port
					port/user-data/config/handler-context/opener port none
				]
			]
			error [print ["*** Client Error:" mold disarm :arg] close port]
		]
	]

	form-http-host: func [
		"Create host name with port number if not the default."
		port
		/local host id
	][
		host: copy port/host
		if all [id: port/port-id id <> 80] [repend host [#":" id]]
		host
	]

	mold-http-target: func [
		"Create path and target on host server (escaped if needed)."
		port
		/local target
	][
		target: to-url "/"
		if port/path [append target port/path]
		if port/target [append target port/target]
		mold target ; escape chars, return result
	]

	make-http-request: func [
		"Add an HTTP header to the request packet."
		port
		data "The request packet - modified by this function"
	][
		head insert data reduce [
			"POST " mold-http-target port " HTTP/1.0" crlf
			"Host: " form-http-host port crlf
			"Cache-Control: no-cache" crlf
			"Pragma: no-cache" crlf
			"Content-Length: " length? data crlf
			; keep-alive? !!!
			crlf
			as-string data
			crlf
		]
	]

]
