REBOL [
	Title: "REBOL/Services - Logging and Debugging"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- Logging and Debugging ----------------------------------------------------
;
; The DEBUG function is used for printing debug information as well as for
; generating detailed logs of operation.
;
; You can enable logging or debugging of selected subsystems by adding
; the associated word to the log-watch or debug-watch lists.
;
; The server and client configuration objects can also change these
; values (and the watch values too).
;
; WARNING: logging a lot of events will slow down the service

; This variables are used so we don't need to pass the config object
; each time DEBUG is called.
debug-enable: true
log-enable:   true
log-file:     %log.txt
log-max-line: 120 ; max size of log line (truncated)
log-rotate:   5   ; # of log rotations
log-limit:    5'000'000 ; max log size before rotation
log-interval: 200 ; # of log lines to check on log size
log-count:    0   ; counter for interval

;-- Events:

debug-watch: [
	system
	config
	log
	error
	test
	service
	connect
	request
;	reply		; long
	packet
;	packet-data	; long
	tcp
	tcp-read
	session
	login
	userbase
]

log-watch: copy debug-watch [
	system
	config
	log
	connect
]

init-log: func [
	"Configure logging and debugging (after config processed)"
	config [object! none!]
	/local val
][
	if not config [exit]
	log-file: config/log-file
	log-count: 0
	if all [exists? log-file  (size? log-file) > log-limit] [rotate-logs]

	; If options is TRUE, enable default events. If it is a block
	; enable only those events. If none, use defaults.
	val: config/debug-watch ; optional config field
	if logic? val [debug-enable: val]
	if block? val [debug-enable: true debug-watch: val]
	val: config/log-watch   ; optional config field
	if logic? val [log-enable: val]
	if block? val [log-enable: true log-watch: val]
]

debug: func [
	"Output a debug or log event if enabled."
	'event "event type" ; from above watch list
	info "information" ; only evaluated if event is enabled
][
	if debug-enable [ ; do not use ALL here (for speed)
		if find debug-watch event [
			print ["DEBUG" join event ":" limit-str remold info log-max-line]
		]
	]
	if log-enable [
		if find log-watch event [
			info: limit-str trim/lines mold/only reduce [now event reduce info] log-max-line
			attempt [write/append log-file append info newline]
			log-count: log-count + 1
			if all [
				log-count > log-interval
				(size? log-file) > log-limit
			][
				log-count: 0
				rotate-logs
			]
		]
	]
]

rotate-logs: func [
	"Rotate log files according to settings"
	/local f1 f2
][
	attempt [delete to-log-file log-rotate]
	for n log-rotate 2 -1 [
		f1: to-log-file n - 1
		f2: to-log-file n
		if exists? f1 [attempt [rename f1 f2]]
	]
	attempt [rename log-file f1]
	debug log ['rotated log-file]
]

to-log-file: func [
	"Generate rotated log file name, eg. log2.txt"
	num
	/local file
][
	file: copy log-file
	head insert any [find/last file "." tail file] num
]

dump-error: func [
	error
][
	write/append %error.txt mold error
]
