REBOL [
	Title: "REBOL/Services - Packet Encoding"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- Packet Handling ----------------------------------------------------------
;   
;   Packets are used for requests (to service) and replies (from service).
;   Each packet has three parts:
;   
;   1. Packet Header 
;   
;   This identifies the packet and provides length and session info. This
;   part of the packet is not encrypted, and it does not contain private
;   information. This header is a null terminated REBOL block with a
;   structure that is verified by PARSE rules (see LOAD-HEADER below for
;   precise structure). The header can indicate that the rest
;   of the packet is encoded data, such as the type of encryption used.
;
;       [rebser version session-id encoding request-len payload-len]
;   
;   2. Request Body
;   
;   This is a REBOL source block of that has a well defined structure
;   used for requests to and replies from services. This is the dialect
;   that makes services work.
;   
;   The raw string length of the body is provided in the header, and the
;   body is separately encrypted to allow it to be decrypted and loaded
;   before the final payload (below) has arrived. This allows applications
;   and tools to provide detailed progress information to the user if that
;   is necessary.
;   
;   3. Payload Area (optional)
;   
;   This is an optional section that holds large strings (and binary)
;   that were removed from the above request block. This allows binary to
;   be be sent as binary rather than as base-64 encoded binary within the
;   request block. This method also allows apps to show progress
;   information before the packet is completely transferred.
;
;   Note:
;   
;   Note that all of the above sections are length-limited for security
;   reasons. If they exceed the configured length, an error will be
;   thrown.
;
;   Regression test is: tests/packet.r

rs-version: 1 ; REBOL Services Packet Version/Compatibility

load-packet: func [
	"Loads rebser packet header and request. Returns block (request), word (error), none (need more)."
	session [object!]
	/local buf header request data num
	; Errors: bad-header, long-header, bad-request, long-request
][
	; This function gets called multiple times until the packet is
	; complete or an error has occurred. This function will return
	; the decoded request block when it has arrived and been decoded.
	; Otherwise, this function returns NONE if more data is still needed,
	; or WORD if an error occurred.

	; To setup progress feedback (for GUI's and tools) check session/request.
	; If it is a block, that info can be displayed, but be aware that the
	; packet has not been expanded (binary pointer pairs are present).

	; See the load-header function for the header block format.

	buf: session/recv-buf

	;-- Find and validate packet header:
	if not header: session/header [
		session/request: none
		either all [
			header: find buf 'rebser  ; skip leading garbage
			data: find/tail header null ; must be terminated
		][
			; Header found, validate it and remove it.
			header: load-header header
			debug packet ['header-in mold header]
			if not header [return error-out session bad-header]
			if rs-version <> ph-version? header [return error-out session bad-version]
			remove/part buf data ; remove header bytes
			session/header: header ; used as a flag that header is ok
		][
			; If header is too large, return error. (Connection should be closed
			; and client should resend the request again.)
			if greater? length? buf session/max-header [
				return error-out session long-header
			]
		]
	]

	;-- Decode (decrypt) request block if it's there:
	if all [
		header
		not request: session/request
	][
		num: ph-req-len? header
		either (length? buf) >= num [
			request: load-request session copy/part buf num
			debug packet ['request-size all [request length? request]]
			if not request [return error-out session bad-request]
			remove/part buf num
			session/request: request
			debug packet-data ['request-data request]
			if not valid-session? session [
				return error-out session bad-session
			]
		][
			; If request is too large, return error:
			if greater? length? buf session/max-request [
				return error-out session long-request
			]
		]
	]

	;-- Decode (decrypt) payload data and insert into request block:
	if request [
		num: ph-pay-len? header
		either (length? buf) >= num [
			; Process request payload if there is one:
			if not zero? num [
				data: decode-packet session copy/part buf num
				debug packet ['payload all [data length? data]]
				if not data [return error-out session bad-payload]
				expand-packet request data
				remove/part buf num
			]
			session/header: none
			return request
		][
			; If payload is too large, return error:
			if greater? length? buf session/max-payload [
				return error-out session long-payload
			]
		]
	] ; else returns NONE
]

error-out: func [
	"Error happened. Reset packet state. Log the error."
	session 'reason
][
	session/header: none ; resets load state
	debug packet ['error reason]
	return reason
]

; Packet header values (keep in-sync with load-header and mold-packet):
ph-version?: :second
ph-session?: :third
ph-encode?:  :fourth
ph-req-len?: :fifth
ph-pay-len?: :sixth

load-header: func [
	"Loads and validates request header, returns packet spec block."
	data
][
	all [
		data: attempt [load/all as-string data]
		parse data [
			'rebser     ; REBOL/Services packet header signature
			integer!    ; version/compatiblity (=1)
			integer!    ; session id (user, encryption, sequence)
			integer!    ; encoding method
			integer!    ; request length
			integer!    ; payload length
			to end
		]
		data
	]
]

header-found?: func [
	"Lower level packet header check. Return session if found. (For CGI)"
	data
	/local header
][
	if all [
		find data null
		header: load-header data
	][
		ph-session? header ; return session #
	]
]

mold-packet: func [
	"Mold a request (or result) packet into standard service format."
	session [object!]
	request [block!]
	/local data packet
][
	data: make binary! 10 ; just a start
	request: compand-packet request data ; modifies both arguments
	; Create, encrypt, and encode the request part of the packet:
	request: append mold/all/only request null
	request: encode-packet session as-binary request
	; Create, encrypt, and encode the payload, if there is one:
	if not empty? data [data: encode-packet session data]
	; Create the proper packet header (see load-header above):
	packet: as-binary reform [
		'REBSer
		1
		session/id
		session/encode
		length? request
		length? data
	]
	debug packet ['header-out as-string packet]
	repend tail packet [null request data]
	packet
]

load-request: func [
	"Decodes/decrypts request. Returns NONE on failure."
	session [object!]
	data
][
	if all [
		data: decode-packet session data
		find/tail data null
	][
		return attempt [load/all as-string data]
	] ; else return NONE
]

compand-packet: func [
	"Move large data elements from the request blocks into the data section."
	request
	data
	/local pos val out
][
	; Breadth-first construction of data section...
	out: make block! length? request
	foreach val request [
		any [
			all [
				any-string? val
				greater? length? val 255 ; only move long values
				pos: length? data
				append data val
				insert insert tail out as-pair pos length? val type?/word val
			]
			all [
				pair? val
				insert insert tail out val 'pair!
			]
			insert/only tail out :val
		]
	]
	forall out [
		if block? val: first out [
			change/only out compand-packet val data
		]
	]
	head out
]

expand-packet: func [
	"Copies data sections into request pair! slots (reverse of compand)."
	request
	data
	/local val type
][
	forall request [
		val: first request
		type: pick request 2 ; there may not be one
		if pair? val [
			if type <> 'pair! [
				val: copy/part skip data val/1 val/2
				if type <> 'binary! [
					val: as-string val
					if type <> 'string! [
						val: make get type val
					]
				]
			]
			change request val
			remove next request
		]
		if block? val [expand-packet val data]
	]
]

; Hmm? Will not work for async-modes TCP port, nor for CGI... -Gabriele

send-packet: func [
	"Send a request or result packet thru a port."
	port [port!]
	request [block!]
	/local session ; Client or server session object
][
	session: port/user-data
	request: mold-packet session request
	debug tcp ['send-packet length? request]
	insert port request
]

deny-ip?: func [
	"Check if IP is allowed or denied."
	config [object!]
	ip
	/local blk
][
	if not tuple? ip [ip: to-tuple ip]
	blk: config/ip-allow
	; If IP is allowed, return false:
	if block? blk [
		forall blk [
			if blk/1 = ip [return false]
			if blk/2 = 'mask [
				if (ip and blk/3) = blk/1 [return false]
				blk: skip blk 2
			]
		]
	]
	blk: config/ip-deny
	; If deny ALL others, then return true:
	if blk = 'all [return true]
	; Otherwise, check if it is explicitly denied:
	if block?  [
		forall blk [
			if blk/1 = ip [return true]
			if blk/2 = 'mask [
				if (ip and blk/3) = blk/1 [return true]
				blk: skip blk 2
			]
		]
	]
	return false
]
