REBOL [
	Title: "REBOL/Services - Service Management"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- Services -----------------------------------------------------------------

default-services: [] ; list of default services [name block ...]

service-obj: context [
	name: 'default
	title: "Default Service"
	description: "An empty service."
	dispatch: func [session command /local args result] [
		debug service ['dispatch name command]
		if not rules [
			rules: make-service-rules commands
			; Bind the rules block within this function to allow
			; access to session, args, and result variables.
			bind rules 'args
		]
		if not parse command rules [throw 'bad-command]
		result
	]
	rules: none  ; dialect block created from command block
	allow: none  ; block of user permissions for service (as a whole)
	commands: [] ; command dialect with comments and productions
]

add-default-service: func [
	"Add an uninitialized built-in service block"
	service [block!]
	/local name
][
	name: select service to-set-word 'name
	if name [
		repend default-services [name service]
	]
]

find-service: func [
	"Return the service object for a given name, or NONE."
	config [object!]
	name [word!]
][
	select config/service-list name
]

set-service: func [
	"Set the current service focus for handling commands. Returns TRUE if set."
	session [object!]
	name [word! none!]
][
	session/this-service: none
	if name [
		if not find-service session/config name [return false]
		session/this-service: name
	]
	true
]

bad-service: func [
	"Throw an error to report that the service is not known."
	name
][
	do-error ["Unknown service:" name]
]

handle-service: func [
	"Add a subservice handler to the list. Return service object."
	config [object!]
	spec [block!] "Object that implements handler"
][
	spec: make service-obj spec
	if not find config/services spec/name [insert config/services spec/name]
	debug service ['handle-service spec/name spec/title]
	insert config/service-list reduce [spec/name spec]
	spec
]

add-service-names: func [
	"Adds default services to service name list"
	config
	names [block!]
	/local sers word
][
	sers: config/services
	foreach name names [
		word: to-word join "-" name
		if all [
			not find sers name
			not find sers word
		][
			insert sers name
		]
	]
] 

load-services: func [
	"Create services from a list of service or file names"
	config [object!]
	/local file-exists? dir data file
][
	dir: any [config/service-dir %services/]
	add-service-names config [file admin home] ; home must be last
	foreach service config/services [
		file: dir/:service
		if none? suffix? file [append file ".r"]
		case [
			any [
				all [
					file-exists?: exists? file
					block? data: attempt [load file]
				]
				data: select default-services service
			] [
				handle-service config data
			]
			file-exists? [
				debug service ['cannot-load file]
			]
			'else [
				debug service ['does-not-exist file]
			]
		]
	]
]

make-service-rules: func [
	"Create parse rules from main command dialect"
	commands [block!]
	/local val blk
][
	blk: make block! (length? commands) / 2 ; estimate
	parse commands [
		some [
			val:
			  #doc block! ; ignore doc block
			| #need [word! | block!]
				(append/only blk to-paren compose/only [need-permission session (val/2)])
			| string! ; ignore doc string
			| block! (append/only blk make-service-rules first val)
			| skip (append/only blk first val)
		]
	]
	blk
]

need-permission: func [
	"Verify user has the required permission (capability)."
	session [object!]
	'needs [word! block!] "e.g. user, admin, editor"
	/local user
][
	user: find-uid session/user-id
	if not user [throw 'need-permission]
	if not block? needs [needs: reduce [needs]]
	foreach tag needs [
		if not any [
			tag = 'user
			user-tag? user tag
		][throw 'need-permission]
	]
]

make-service-doc: func [
	"Create language-specific dialect summary from main dialect."
	commands [block!]
	lang [word!]
	/local val blk
][
	blk: make block! (length? commands) / 2 ; estimate
	parse commands [
		some [
			val:
			#doc set val block! (
				val: any [select val lang val/2] ; first is default
				if val [append blk val]
			)
			| paren! ; remove production
			| set-word!
			| block! (append/only blk make-service-doc first val lang)
			| skip (append/only blk first val)
		]
	]
	new-line/all blk off
	forall blk [
		if equal? '| pick blk 1 [new-line blk on]
	]
	blk
]

