REBOL [
	Title: "REBOL/Services - Server User Database"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- User Database Record -----------------------------------------------------
;
;	The user database is a set of blocks that hold info for each user.
;	This method will work fine for supporting up to a few thousand users.
;	The data is stored as a REBOL file and is usually loaded into memory full
;	time while the service is running. When the file is loaded it is
;	parsed to validate all fields. This is a check, just in case the
;	file was edited by the user manually outside of this code.
;
;	Once a user has logged onto the system with their user login name, they
;	are referred to by id number, not by name. This allows the user to change
;	their name if that is necessary and not affect anything else. This user
;	id is also used within any other records as identification of the user
;	(as is done in relational databases).
;
;	A special "contact" field is provided for custom information used to
;	identify or reach the user (such as a email address or employee number).
;	If you need to add additional information for each user (location, profile,
;	status, etc.) add it to a separate file (and use the id to relate it back
;	to this user-base). Doing so will minimize the activity (load/save) needed
;	for this user-base and will also keep in-memory information to a minimum.

user-record: [
	id		integer!			; user number (unique - can be used to ref other DB)
	user	string!				; user login name
	name	string!				; full name
	tags	block!				; block of flags & permissions
	auth	binary!				; hashed password
	salt	binary!				; hash key
	create	date!				; date user account was created
	modify	date!				; date account was last modified
	expire	[date! | integer!]	; expires on or logon count (0 for always)
	contact	binary!				; user contact info (email, phone, employee #, etc.)
]

private-fields: [auth salt] ; do not allow these to be queried

;-- Special Userbase Variables -----------------------------------------------

user-file:		%userbase.rdb	; user database

user-log:		%log-user.txt	; user database change log

user-base:		[]				; block of user records
user-fields:	[]				; list of user field names
user-valid:		[]				; parse rule for user record

;-- Userbase Initialization, Load, and Save ----------------------------------

do init-user-fields: func [
	"Enumerate user field values, construct parser validator for records."
	/local num word
][
	clear user-fields
	clear user-valid
	num: 0
	foreach [field type] user-record [
		append user-fields field
		append/only user-valid type
		word: to-word join 'user- field
		set word num: num + 1
	]
	new-line/all user-fields off
]

user-base-header: [REBOL [
	Title: "REBOL/Services User Database"
	Version: 1.0.0
	Date: (now)
	Fields: (user-fields)
]]

save-userbase: func [
	"Save userbase to disk file."
	/local out err
][
	out: append compose/deep/only user-base-header user-base
	foreach user skip out 2 [new-line/all user off]
	new-line/all skip out 2 on
	if error? err: try [write user-file mold/only out 0][
		rs-error 'user-base 'save disarm :err
	]
]

load-userbase: func [
	"Load userbase from a disk file. Return NONE if it does not exist."
	file "Userbase file name"
	/local data spot
][
	debug userbase ['load-userbase file exists? file]
	user-file: file

	; If the user database does not exist, auto-create admin user:
	if not exists? user-file [
		create-user/tags "admin" "Administrator" "pass" [admin]
		save-userbase
		return user-base
	]
	if error? data: try [load/all/header user-file][
		rs-error 'user-base 'load disarm :data
	]
	if not parse data [
		object!
		any [spot: into user-valid]
	][
		print ["parse error:" spot]
		rs-error 'user-base 'parse spot
	]
	if not find/match user-fields data/1/fields [
		print ["format error:" user-fields]
		rs-error 'user-base 'format data
	]
	remove user-base: data
	data
]

;-- User Journal -------------------------------------------------------------
;
;	Implements a journal for tracking important user events like creation,
;	deletion, and account changes. This is kept as a loadable REBOL file.

note-user: func [
	"Make a note on an important user event"
	event "Specific event"
	info  "Additional information"
][
	info: new-line/all reduce info off
	info: append mold/only reduce [now event info] newline
	attempt [write/append user-log info]
]

;-- User Management Functions ------------------------------------------------
;
;	Adds and removes users from database.
;
;	Note that these functions do not save the changes in order to allow
;	batch changes of multiple user accounts with a single final save.
;
;	An applications's higher level functions must call the save-userbase
;	function when all changes are done.

create-user: func [
	"Create a new user, return user record or NONE if user exists."
	user [string!]
	name [string!]
	pass [string! binary!]
	/tags tag [block!]
	/contact cont [email! string!]
	/expires expire [date! integer!]
	/local id auth salt create modify
][
	if any [empty? user find-user user] [return none]
	note-user 'create [id user name]
	id: either empty? user-base [1] [1 + first last user-base]
	tags: copy any [tag []]
	salt: make-salt
	contact: to-binary any [cont ""]
	auth: make-pass pass salt
	create: modify: now
	expire: any [expire 0]
	user: reduce bind user-fields 'local ; global block bind is ok
	append/only user-base user
	user ; return.
	; (call save-userbase function to make it permanent)
]

delete-user: func [
	"Delete a user from database."
	uid [integer!]
	/local user
][
	if not user: find-uid uid [exit]
	remove-each user user-base [uid = pick user user-id]
	note-user 'delete copy/part user 3
	; (call save-userbase function to make it permanent)
]

change-user: func [
	"Change specified user fields. (Do not use for id/auth/salt data.)"
	uid [integer!]
	fields [block!] "Block of [field value ...]"
	/local user
][
	if not user: find-uid uid [exit]
	if find fields 'id [exit] ; not allowed
	foreach [field value] fields [
		if all [
			i: find user-fields field
			i: index? i
			parse reduce [value] reduce [pick user-valid i]
		][
			put user i value
		]
	]
	put user user-modify now
	note-user 'change copy fields
	; (call save-userbase function to make it permanent)
]

change-pass: func [
	"Change user password."
	uid [integer!]
	pass [string!]
	/local user salt
][
	if not user: find-uid uid [exit]
	salt: make-salt
	put user user-salt salt
	put user user-auth make-pass pass salt
	put user user-modify now
	note-user 'pass copy/part user 2
	; (call save-userbase function to make it permanent)
]

;-- Search and Query Functions -----------------------------------------------

find-uid: func [
	"Returns user record from user id."
	uid [integer!]
][
	foreach user user-base [
		if uid = pick user user-id [return user]
	]
	none
]

find-user: func [
	"Returns user record from user name."
	name [string!]
][
	foreach user user-base [
		if name = pick user user-user [return user]
	]
	none
]

list-users: func [
	"Returns a block of specified user fields."
	fields [block! word! none!] "Fields being requested"
	/local users flag
][
	either none? fields [
		; Get all non-private fields:
		fields: [id user name tags create modify expire contact]
		flag: true
	][
		if word? fields [fields: reduce [fields]]
		fields: exclude fields private-fields ; must remain secure
	]
	users: make block! (length? user-base) * length? fields
	foreach user user-base [
		user: map-words user-fields fields user
		either flag [
			append/only users user
		][
			append users user
		]
	]
	users
]


;-- Password Challenge Authentication ----------------------------------------

authenticate-user?: func [
	"Authenticate a challenge response (server side). True or false."
	uid [integer!]
	challenge [binary!]
	response [binary!]
][
	if all [
		uid: find-uid uid
		able-user? uid
		response = pass-challenge pick uid user-auth challenge
	][
		note-user 'auth [pick uid user-id pick uid user-user]
		true
	]
]

make-challenge: func [
	"Given a user id, returns [challenge user-salt]."
	uid [integer!]
	/local salt
][
	salt: either uid: find-uid uid [pick uid user-salt][make-salt]
	reduce [make-salt salt] ; [challenge user-salt]
]


;-- Attribute Related Functions ----------------------------------------------
;
;	Each user account allows a block of tags that can be sent in two ways.
;	You can just include a word to indicate a tag exists, or you an give
;	a word followed by a value. It is up to the user to know if a specific
;	tag includes an additional value.

set-user-tag: func [
	"Set a tag and optional value in the user's account."
	user [block!]
	tag [word! block!]
	/delete "Delete the tag"
	/local tags
][
	tags: pick user user-tags
	either block? tag [
		remove/part find tags tag/1 2
	][
		remove find tags tag
	]
	if not delete [append tags tag]
	note-user 'tag [pick user user-id pick user user-name del tag]
]

user-tag?: func [
	"Return tag value if user account has the specified tag set."
	user [block!]
	tag [word!]
][
	user: pick user user-tags
	if user: find user tag [
		any [user/2 true] ; return next value or TRUE if at end
	]
]

disable-user: func [
	"Suspend the user's account."
	uid [integer! block!]
][
	if integer? uid [if not uid: find-uid uid [exit]]
	set-user-tag uid 'disable
]

enable-user: func [
	"Activate the user's account."
	uid [integer! block!]
][
	if integer? uid [if not uid: find-uid uid [exit]]
	set-user-tag/delete uid 'disable
]

able-user?: func [
	"Returns TRUE if user account is enabled."
	uid [integer! block!]
][
	if integer? uid [if not uid: find-uid uid [return false]]
	not user-tag? uid 'disable
]

admin-user?: func [
	"Returns TRUE if user is an admin."
	uid [integer! block!]
][
	if integer? uid [if not uid: find-uid uid [return false]]
	user-tag? uid 'admin
]

auth-user?: func [
	"Return user account auth (pass hash)."
	uid [integer!]
][
	all [
		uid: find-uid uid
		able-user? uid
		pick uid user-auth
	]
]
